package de.opalpiraten.scenepainter.gui.render;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

/**
 * The Scene is the component on which the painting takes place. It is
 * organized in layers of SceneElements (one per layer).
 */
public class Scene extends JPanel implements MouseMotionListener, 
MouseListener, MouseWheelListener {

  /**
   * Set to true every time, the scene gets changed. This flag is intended
   * to be used for synchronizing the scene. That is, whenever the
   * Scene is written out to disk, the writer should reset it to false again.
   */
  public boolean dirty = false;
  
  /**
   * Mark selected SceneElement in this color.
   */
  private Color selectedcolor=Color.BLUE;
  
  /**
   * Stores the elements. Top layer is at index 0.
   */ 
  private SceneElement[] elements = new SceneElement[0];
  
  /**
   * The Grid to draw below the SceneElements. 
   */
  private Grid grid = null;
  
  /**
   * Bounds of the clip area (for exporting subimages)
   */
  private Rectangle clipping;
  
  /**
   * The index of the selected layer (layer to work on). If negative, no
   * layer is selected.
   */
  private int selected=-1;
  
  /**
   * The SceneElement, that is currently being dragged (null if none is 
   * dragged).
   */
  private SceneElement dragging;
  
  /**
   * Difference between the position of the mousecursor and the 
   * SceneElement origin (used for dragging).
   */
  private Point offset;
  
  /**
   * The SceneListeners registered with this Scene
   */
  private Vector listeners = new Vector();
  
  /**
   * Whether or not SceneElements should snap to the grid
   * (true==snap to grid).
   */
  private boolean snapgrid = false;
  
  /**
   * If the grid should be drawn on top or below the layerstack
   */
  private boolean ontopgrid = false;
  
  
  /**
   * To get a bit more responsiveness from the UI when dragging big 
   * images around: This flag is set to true while the component
   * repaints. While true, no other repaint should be started
   */
  private boolean painting=false;
  
  
  public Scene () {
    super(false);
    addMouseMotionListener(this);
    addMouseListener(this);
    addMouseWheelListener(this);
  }
  
  /**
   * Query the preferred  size for this Scene to show all elements.
   * @return The expansion of the scene calculated from all elements with
   * positive coordinates. The minimumsize is always 1:1.
   */
  public Dimension getPreferredSize() {
    Dimension ret = new Dimension(1,1);
    for (int i=0;i<elements.length;i++) {
      Rectangle r = elements[i].getBounds();
      if (r.x+r.width>ret.width) ret.width=r.x+r.width;
      if (r.y+r.height>ret.height) ret.height=r.y+r.height;
    }
    return ret;
  }
  
  /**
   * Set the object responsible for drawing the Grid
   * @param g the grid
   */
  public void setGrid(Grid g) {
    grid=g;
    repaint();
  }
  
  /**
   * Query the Grid
   */
  public Grid getGrid() {
    return grid;
  }
  
  /**
   * Add a SceneElement to this Scene, The new element will be placed on a
   * new layer on top of the other layers and the scene will be redrawn.
   * Adding an element, that is already part of the scene has no effect.
   */
  public void add(SceneElement elem) {
    for (int i=0;i<elements.length;i++) {
      if (elements[i]==elem) {
        return;
      }
    }

    SceneElement[] tmp = new SceneElement[elements.length+1];
    System.arraycopy(elements,0,tmp,1,elements.length);
    tmp[0]=elem;
    elements=tmp;
    fireSceneChanged();
  }
  
  /**
   * Remove a SceneElement from the Scene.
   */
  public void remove(SceneElement elem) {
    for (int i=0;i<elements.length;i++) {
      if (elements[i]==elem) {
        SceneElement[] tmp = new SceneElement[elements.length-1];
        System.arraycopy(elements,0,tmp,0,i);
        System.arraycopy(elements,i+1,tmp,i,tmp.length-i);
        elements=tmp;
        selected=-1;
        fireSceneChanged();
      }
    }
  }

  /**
   * Clear Scene
   */
  public void removeAll() {
    elements = new SceneElement[0];
    selected=-1;
    fireSceneChanged();
  }
  
  /**
   * Replace a layer. If any contained SceneElements were edited through
   * their setter methods, they must be resubmitted through this method in
   * order to make the change known.
   * @param i layer to replace (nothing happens if out of bounds
   * @param s what to put on the layer.
   */
  public void replace(int i,SceneElement s) {
    if (i<0 || i>=elements.length) return;
    elements[i]=s;
    fireSceneChanged();
  }
  
  /**
   * Raise the selected layer 
   */
  public void lower() {
    if (selected<0 || selected >=elements.length-1) return;
    SceneElement tmp=elements[selected+1];
    elements[selected+1]=elements[selected];
    elements[selected]=tmp;
    selected++;
    
    fireSceneChanged();
  }

  /**
   * Lower the selected layer.
   */
  public void raise() {
    if (selected<1 || selected >=elements.length) return;
    SceneElement tmp=elements[selected-1];
    elements[selected-1]=elements[selected];
    elements[selected]=tmp;
    selected--;
     
    fireSceneChanged();
  }
  
  /**
   * This method must be used to move elements, so SceneListeners are
   * notified.
   */
  public void move(SceneElement e, Point p) {
    e.position=p;
    fireSceneChanged();
  }
  
  /**
   * This method must be used to set the transparency of the selected
   * layer.
   */
  public void setAlpha(int layer, float alpha) {
    if (layer<0 || layer >= elements.length) return;
    if (alpha<0.0f || alpha>1.0f) return;
    elements[layer].setAlpha(alpha);
    repaint();
    fireSceneChanged();
  }
  
  /**
   * Set the policy for drawing the grid
   * @param b if true, the grid is drawn on top of the layerstack. If
   * false, the grid is drawn below.
   */
  public void setOntopgrid(boolean b) {
    ontopgrid=b;
    fireSceneChanged();
  }
  
  /**
   * Query the policy for drawing the Grid
   */
  public boolean isOntopgrid() {
    return ontopgrid;
  }
  
  /**
   * Make an element the selected element. Selected elements are visually
   * marked and SceneListeners are notified.
   * @param i index number of the layer to select
   */
  public void setSelectedIndex(int i) {
    if (i<0 || i>=elements.length) selected=-1;
    else selected = i;
    fireSceneChanged();
  }
  
  /**
   * Query the layernumber of the selected layer
   * @return the indexnumber of the selected layer or -1 if none is selected.
   */
  public int getSelectedIndex() {
    return selected;
  }
  
  /**
   * Query the selected element
   * @return either the element or null if none is selected
   */
  public SceneElement getSelected() {
    if (selected <0) return null;
    return elements[selected];
  }
  
  /**
   * Set the color for marking the selected element
   * @param c either the color or null to not mark the selected element
   */
  public void setSelectedColor(Color c) {
    selectedcolor=c;
    repaint();
  }

  
  /**
   * Query all elements, the Scene consists of
   * @return the elements sorted by their layers. the topmost element
   * is at index 0.
   */
  public SceneElement[] getElements() {
    return elements;
  }
  
  /**
   * Load elements into the scene. This replaces the contents of the 
   * current scene and is intended to be used by the beans API.
   */
  public void setElements(SceneElement[] e) {
    elements = e;
    fireSceneChanged();
  }
  
  /**
   * Request the Scene to snap SceneElements to the Grid
   */
  public void setSnapgrid(boolean b) {
    snapgrid=b;
  }
  
  /**
   * Query whether or not the Scene snaps SceneElements to the Grid
   */
  public boolean isSnapgrid() {
    return snapgrid;
  }
  
  /**
   * Export the Scenes content.
   * @return a "screenshot" of the scene, containing all screenelements having
   * positive coordinates. Elements not having (entirely) positive coordintates
   * are cropped.
   */
  public BufferedImage export() {
    Dimension exp = new Dimension(0,0);
    for (int i=0;i<elements.length;i++) {
      Rectangle eb =elements[i].getBounds();
      if (eb.x+eb.width>exp.width) exp.width=eb.x+eb.width;
      if (eb.y+eb.height>exp.height) exp.height = eb.y+eb.height;
    }
    
    Rectangle bounds = new Rectangle(exp.width,exp.height,0,0);
    for (int i=0;i<elements.length;i++) {
      Rectangle eb =elements[i].getBounds();
      if (eb.x<bounds.x) bounds.x=eb.x;
      if (eb.y<bounds.y) bounds.y=eb.y;
      if (eb.x+eb.width>bounds.width) bounds.width=eb.x+eb.width;
      if (eb.y+eb.height>bounds.height) bounds.height=eb.y+eb.height;
    }
    if (bounds.x<0) bounds.x=0;
    if (bounds.y<0) bounds.y=0;
    bounds.width=bounds.width-bounds.x;
    bounds.height=bounds.height-bounds.y;
    if (bounds.width<1) bounds.width=1;
    if (bounds.height<1) bounds.height=1;
    if (exp.width<1) exp.width=1;
    if (exp.height<1) exp.height=1;
    BufferedImage bi = new BufferedImage(exp.width,exp.height,BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = bi.createGraphics();
    for(int i=elements.length-1;i>=0;i--) elements[i].paint(g);
    return bi.getSubimage(bounds.x,bounds.y,bounds.width,bounds.height);
  }
  
  /**
   * Register with the Scene. Registering twice will have no effect.
   */
  public void addSceneListener(SceneListener sl) {
    if (listeners.contains(sl)) return;
    listeners.add(sl);
  }
  
  /**
   * De-Register with the Scene
   */
  public void removeSceneListener(SceneListener sl) {
    listeners.remove(sl);
  }
  
  /**
   * Notify listeners, that the scene changed
   */
  private void fireSceneChanged() {
    dirty=true;
    repaint();
    /*
    Enumeration e = listeners.elements();
    while (e.hasMoreElements()) {
      SceneListener l = (SceneListener)e.nextElement();
      l.sceneChanged(this);
    }
    */
    int size = listeners.size();
    for (int i=0;i<size;i++)  {
      ((SceneListener)listeners.get(i)).sceneChanged(this);
    }
  }
  
  
  /**
   * Query the layerindex of all elements at a specific point
   * @return all elements, occupying the desired spot, with the bottommost
   * layer at index 0.
   */
  public int[] getIndexesAt(Point p) {
    int tmp[] = new int[elements.length];
    int idx = 0;
    for(int i=0;i<elements.length;i++) {
      if(elements[i].getBounds().contains(p)) {
        tmp[idx]=i;
        idx++;
      }
    }
    int ret[] = new int[idx];
    System.arraycopy(tmp,0,ret,0,ret.length);
    return ret;
  }
  
  /**
   * A Scene may have listeners set, that should not be mapped out
   * @param the scene to strip
   * @return a new scene object only carrying the actual sceneinformation.
   * (shallow copy)
   */
  public static Scene strip (Scene s) {
    Scene ret = new Scene();
    ret.setElements(s.getElements());
    ret.setSelectedIndex(s.getSelectedIndex());
    ret.setName(s.getName());
    ret.setSnapgrid(s.isSnapgrid());
    ret.setOntopgrid(s.isOntopgrid());
    ret.setGrid(s.getGrid());
    ret.setBackground(s.getBackground());
    ret.setSize(s.getSize());
    return ret;
  }

   
  public void paintComponent(Graphics g) {
    painting=true;
    super.paintComponent( g );
    Color tmp = g.getColor();
    if (grid!=null && !ontopgrid) grid.paint((Graphics2D)g,getWidth(),getHeight());
    for (int i=elements.length-1;i>=0;i--) {
      elements[i].paint((Graphics2D)g);
    }
    if (selected!=-1 && selectedcolor!=null) {
      g.setColor(selectedcolor);
      Rectangle r = elements[selected].getBounds();
      g.drawRect(r.x,r.y,r.width,r.height);
    }
    if (grid!=null && ontopgrid) grid.paint((Graphics2D)g,getWidth(),getHeight());
    g.setColor(tmp);
    painting=false;
  }
  
  
  // Interface implementations below
  public void mouseClicked(MouseEvent e) {}
  
  public void mouseEntered(MouseEvent e) {}
  
  public void mouseExited(MouseEvent e) {}
  
  public void mousePressed(MouseEvent e) {
    Point pos = e.getPoint();
    switch(e.getButton()) {
      case MouseEvent.BUTTON1: {
        int tmp[]=getIndexesAt(pos);
        if (tmp.length==0) {
          // User clicked on empty space. If an element is selected, deselect
          // it. Do nothing otherwise.
          if (selected>=0) setSelectedIndex(-1);
          break;
        }
        
        // There may be a stack of elements under the mousecursor and maybe
        // an element is already selected, that is not on top of the stack.
        // in this case  the selected element stays selected and will be
        // dragged. Otherwise the topmost element of the stack becomes the
        // selected and dragged element.
        if (selected>=0 && elements[selected].getBounds().contains(pos)) {
          dragging=elements[selected];
        }
        else {
          dragging=elements[tmp[0]];
          setSelectedIndex(tmp[0]);
        }
        offset=new Point(pos.x-dragging.position.x,pos.y-dragging.position.y);
        break;
      }
      
      case MouseEvent.BUTTON3: {
        int[] tmp = getIndexesAt(pos);
        if(tmp.length==0) break;
        selected=-1;
        remove(elements[tmp[0]]);
        break;
      }
    }
  }
  
  public void mouseReleased(MouseEvent e) {
    if (dragging!=null) {
      dragging=null;
      fireSceneChanged();
    }
  }
  
  public void mouseDragged(MouseEvent e) {
    
    if (dragging!=null) {
      Point orig = dragging.position;
      Point pos = e.getPoint();
      pos.x=pos.x-offset.x;
      pos.y=pos.y-offset.y;
      
      if (snapgrid && grid!=null) {
        dragging.position=grid.getCellOrigin(pos);

        if (e.isAltDown()) {
          for (int i=0;i<elements.length;i++) { 
            if (elements[i]==dragging) continue;
            elements[i].position.x=elements[i].position.x+dragging.position.x-orig.x;
            elements[i].position.y=elements[i].position.y+dragging.position.y-orig.y;
          }
        }
      }
      
      else {
        dragging.position=pos;

        if (e.isAltDown()) {
          for (int i=0;i<elements.length;i++) {
            if (elements[i]==dragging) continue;
            elements[i].position.x=elements[i].position.x+pos.x-orig.x;
            elements[i].position.y=elements[i].position.y+pos.y-orig.y;
          }
        }
      }
      
      // Performance tweak: if a previous event is still painting, don't
      // trigger repainting.
      if (!painting) {
        repaint();
      }
    }
  }

  public void mouseMoved(MouseEvent e) {
    
  }
  
  public void mouseWheelMoved(MouseWheelEvent e) {
    if (e.isAltDown()) {
      if (e.getWheelRotation()>0 ) {
        if (!(selected<0)) elements[selected].scrollView(true);
      }
      else {
        if (!(selected<0)) elements[selected].scrollView(false);
      }
      fireSceneChanged();
      return;
    }
    
    if (e.getWheelRotation()<0 ) {
      lower();
    }
    else {
      raise();
    }
  }
  
  public String toString() {
    return getName();
  }
  
}
